/*
 * This is a simple application for recording audio. The device for
 * recording is specified on the command line, the recording parameters
 * have to be adjusted in this file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * The author is Andreas Bogk <andreas@convergence.de>
 */
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <stdio.h>

#define BUF_SIZE  16384
#define STDOUT    1

int audio_fd;
unsigned char audio_buffer[BUF_SIZE];

int main(int argc, char** argv) {
  int len;
  int param;
  int format = AFMT_S16_LE;    /* 16 bit, signed, litte endian */
  int stereo = 1;              /* 0=mono, 1=stereo */
  int speed  = 44101;          /* sampling rate 44.1 kHz */

  if((audio_fd = open(argv[1],O_RDONLY)) < 0) {
    perror(argv[1]);
    exit(1);
  }

  
  param = format;
  if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &param) == -1)  {
    perror("SNDCTL_DSP_SETFMT");  
    // exit(1);  
  }  
  
  if (format != param) {  
    fprintf(stderr, "Your soundcard doesn't support requested audio mode. Throw it away and get a new one.\n");
    exit(1);
  }  

  param = stereo;
  if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &param)==-1) { 
    perror("SNDCTL_DSP_STEREO");
    // exit(1);
  }
  
  if (stereo != param) {
      fprintf(stderr, "Your soundcard doesn't support the required number of channels. Throw it away and get a new one.\n");
      exit(1);
  }

  param = speed;
  if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &param)==-1) {
    perror("SNDCTL_DSP_SPEED");  
    // exit(1);  
  }  

  if(param != speed) {
      fprintf(stderr, "Your soundcard doesn't support the required samplerate. Throw it away and get a new one.\n");
      fprintf(stderr, " wanted %d got %d\n",speed,param);
/*      exit(1); */
  }

  for(;;) {
    if((len = read(audio_fd, audio_buffer, BUF_SIZE)) == -1) {
      perror("audio_read");
      exit(1); 
    }
    
    if(write(STDOUT, audio_buffer, len) < len) {
      perror("write");
      exit(1);
    } 
  } 

}
